#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from future import standard_library

standard_library.install_aliases()
import queue

from alert_utils.alert_utils_common.metric_collector.event_writer import *


class MemoryEventWriter(MetricEventWriter):
    """
    This class is used to mock a in memory event sink.
    Be careful: all the events are stored in the memory queue
    """

    def __init__(self, app, config):
        super().__init__(app, config)
        max_queue_size = config.get("max_queue_size", 0)
        self.q = queue.Queue(maxsize=max_queue_size)

    def _flush_msg(self, msg):
        self.q.put(msg)

    def pop_msg(self):
        return self.q.get()

    def get_msg_count(self):
        return self.q.qsize()



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import threading

from future.utils import with_metaclass
from solnlib import log, pattern

from . import event_writer, memory_event_writer
from .metric_exception import MetricException
from .number_metric_collector import NumberMetricCollector

__all__ = ["Monitor"]

logger = log.Logs().get_logger("metric_collector")


class Monitor(with_metaclass(pattern.Singleton, object)):
    EVENT_WRITERS = {
        "memory": memory_event_writer.MemoryEventWriter,
        "file": event_writer.FileEventWriter,
        "hec": event_writer.FileEventWriter,  # TODO; implement a HEC writer
    }

    def __init__(self):
        self._app = None
        self._ewriter = None
        self._ewriter_type = None
        self.number_metric_collector = None
        self.worker_start_lock = threading.Lock()

    def configure(self, conf, force_update=False):
        """
        conf is a dict.
         -- app: the app name
         -- event_writer: the type of event writer.
         -- writer_config: a dict which contains the configuration for the event writer
        force_update: when this is True, update the config of monitor
        """
        if self.number_metric_collector is not None and force_update is False:
            logger.info("monitor has been initialized.")
            return self

        self._app = conf.get("app", self._app)
        if not self._app:
            raise MetricException("app is not found in configuration object.")

        event_writer = str.lower(conf.get("event_writer", "file"))

        if self._ewriter_type != event_writer:
            if event_writer not in self.EVENT_WRITERS:
                raise MetricException("Unknown event writer type:" + event_writer)
            ewriter_cls = self.EVENT_WRITERS[event_writer]
            writer_config = conf.get("writer_config", {})
            self._ewriter_type = event_writer
            self._ewriter = ewriter_cls(self._app, writer_config)
        else:
            writer_config = conf.get("writer_config", {})
            if self._ewriter:
                self._ewriter.update_config(writer_config)
            else:
                raise MetricException(
                    "event writer type is "
                    + self._ewriter_type
                    + ", while event writer is None."
                )

        if self.number_metric_collector is None:
            self.number_metric_collector = NumberMetricCollector(self._ewriter)
        else:
            self.number_metric_collector.set_event_writer(self._ewriter)
        logger.info(
            "Initialize monitor successfully. app=%s, config:%s", self._app, conf
        )
        return self

    def start(self):
        with self.worker_start_lock:
            if self.number_metric_collector:
                if self.number_metric_collector.is_stopped():
                    self.number_metric_collector.start()
                else:
                    logger.info("Collector thread has been started.")
            else:
                raise ValueError("Monitor is not configued yet.")

    def stop(self):
        if self.number_metric_collector:
            self.number_metric_collector.stop()
            self.number_metric_collector.join()

    def is_stopped(self):
        if self.number_metric_collector:
            return self.number_metric_collector.is_stopped()
        else:
            return True

    def write_event(self, ev, tags=[]):
        if self._ewriter:
            self._ewriter.write_event(ev, tags)

    def register_metric(
        self,
        metric_name,
        metric_type=NumberMetricCollector.SUM_METRIC,
        metric_tags=[],
        max_time_span=10,
    ):
        self.number_metric_collector.register_metric(
            metric_name, metric_type, metric_tags, max_time_span
        )

    def record_metric(self, metric_name, metric_value, metric_timestamp=None):
        if self.number_metric_collector:
            self.number_metric_collector.record_metric(
                metric_name, metric_value, metric_timestamp
            )



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import os
from json import loads as jloads
from os import linesep
from os import path as op

from mako import exceptions
from mako.template import Template
from munch import Munch

from . import alert_actions_exceptions as aae
from . import arf_consts as ac
from .alert_actions_helper import write_file
from .alert_actions_merge import remove_alert_from_conf_file
from .alert_actions_template import AlertActionsTemplateMgr


class AlertActionsConfBase:
    def __init__(
        self,
        input_setting=None,
        package_path=None,
        logger=None,
        template_dir=None,
        default_settings_file=None,
        global_settings=None,
        **kwargs
    ):
        self._alert_conf_name = "alert_actions.conf"
        self._alert_spec_name = "alert_actions.conf.spec"
        self._eventtypes_conf = "eventtypes.conf"
        self._tags_conf = "tags.conf"
        self._app_conf = "app.conf"
        self._all_settings = input_setting
        self._alert_settings = input_setting[ac.MODULAR_ALERTS]
        self._package_path = package_path
        self._logger = logger
        self._global_settings = global_settings

    def get_local_conf_file_path(self, conf_name=None, create_dir_path=True):
        if not self._package_path:
            return None

        if not conf_name:
            conf_name = self._alert_conf_name

        local_path = op.join(self._package_path, "default")
        if not op.exists(local_path) and create_dir_path:
            os.makedirs(local_path)

        return op.join(local_path, conf_name)

    def get_spec_file_path(self, spec_file=None, create_dir_path=True):
        if not self._package_path:
            return None

        if not spec_file:
            spec_file = self._alert_spec_name

        readme_path = op.join(self._package_path, "README")
        if not op.exists(readme_path) and create_dir_path:
            os.makedirs(readme_path)
        return op.join(readme_path, spec_file)

    def get_icon_dir(self, create_dir_path=True):
        icon_dir = op.join(self._package_path, "appserver", "static")
        if not op.exists(icon_dir) and create_dir_path:
            os.makedirs(icon_dir)
        return icon_dir

    def get_icon_name(self, alert):
        return "alert_" + alert[ac.SHORT_NAME] + ".png"

    def get_icon_path(self, alert, create_dir_path=True):
        return op.join(
            self.get_icon_dir(create_dir_path=create_dir_path),
            self.get_icon_name(alert),
        )


class AlertActionsConfGeneration(AlertActionsConfBase):
    DEFAULT_CONF_TEMPLATE = "alert_actions.conf.template"
    DEFAULT_SPEC_TEMPLATE = "alert_actions.conf.spec.template"
    DEFAULT_SETTINGS_FILE = "alert_actions_conf_default_settings.json"
    DEFAULT_ALERT_ICON = "alerticon.png"
    DEFAULT_EVENTTYPES_TEMPLATE = "eventtypes.conf.template"
    DEFAULT_TAGS_TEMPLATE = "tags.conf.template"
    DEFAULT_APP_TEMPLATE = "app.conf.template"

    def __init__(
        self,
        input_setting=None,
        package_path=None,
        logger=None,
        template_dir=None,
        default_settings_file=None,
        **kwargs
    ):
        if not input_setting or not logger:
            msg = 'status="failed", required_args="input_setting, logger"'
            raise aae.AlertActionsInValidArgs(msg)

        super().__init__(
            input_setting=input_setting,
            package_path=package_path,
            logger=logger,
            template_dir=template_dir,
            default_settings_file=default_settings_file,
            **kwargs
        )

        self._html_fields = [
            ac.PARAMETERS,
        ]
        self._remove_fields = [ac.SHORT_NAME] + self._html_fields
        self._temp_obj = AlertActionsTemplateMgr(template_dir)
        self._default_settings_file = default_settings_file or op.join(
            op.dirname(op.abspath(__file__)),
            AlertActionsConfGeneration.DEFAULT_SETTINGS_FILE,
        )

    def generate_conf(self):
        self._logger.info(
            'status="starting", operation="generate", '
            + 'object="alert_actions.conf", object_type="file"'
        )
        template = Template(
            filename=op.join(
                self._temp_obj.get_template_dir(),
                AlertActionsConfGeneration.DEFAULT_CONF_TEMPLATE,
            )
        )
        alert_obj = Munch.fromDict(self._alert_settings)
        try:
            final_string = template.render(mod_alerts=alert_obj)
        except:
            print(exceptions.html_error_template().render())
            raise
        text = linesep.join([s.strip() for s in final_string.splitlines()])
        write_file(
            self._alert_conf_name, self.get_local_conf_file_path(), text, self._logger
        )
        self._logger.info(
            'status="success", operation="generate", '
            + 'object="alert_actions.conf", object_type="file"'
        )

    def generate_eventtypes(self):
        self._logger.info(
            'status="starting", operation="generate", '
            + 'object="eventtypes.conf", object_type="file"'
        )
        template = Template(
            filename=op.join(
                self._temp_obj.get_template_dir(),
                AlertActionsConfGeneration.DEFAULT_EVENTTYPES_TEMPLATE,
            )
        )
        alert_obj = Munch.fromDict(self._alert_settings)
        final_string = template.render(mod_alerts=alert_obj)
        text = linesep.join([s.strip() for s in final_string.splitlines()])
        file_path = self.get_local_conf_file_path(conf_name=self._eventtypes_conf)
        write_file(self._eventtypes_conf, file_path, text, self._logger)

        # remove the stanza if not checked
        for alert in self._alert_settings:
            if alert.get("active_response") and alert["active_response"].get(
                "sourcetype"
            ):
                continue
            remove_alert_from_conf_file(alert, file_path, self._logger)
        self._logger.info(
            'status="success", operation="generate", '
            + 'object="eventtypes.conf", object_type="file"'
        )

    def generate_tags(self):
        self._logger.info(
            'status="starting", operation="generate", '
            + 'object="tags.conf", object_type="file"'
        )
        template = Template(
            filename=op.join(
                self._temp_obj.get_template_dir(),
                AlertActionsConfGeneration.DEFAULT_TAGS_TEMPLATE,
            )
        )
        alert_obj = Munch.fromDict(self._alert_settings)
        final_string = template.render(mod_alerts=alert_obj)
        text = linesep.join([s.strip() for s in final_string.splitlines()])
        file_path = self.get_local_conf_file_path(conf_name=self._tags_conf)
        write_file(self._tags_conf, file_path, text, self._logger)

        # remove the stanza if not checked
        for alert in self._alert_settings:
            if alert.get("active_response") and alert["active_response"].get(
                "sourcetype"
            ):
                continue
            remove_alert_from_conf_file(alert, file_path, self._logger)
        self._logger.info(
            'status="success", operation="generate", '
            + 'object="tags.conf", object_type="file"'
        )

    def generate_spec(self):
        self._logger.info(
            'status="starting", operation="generate", '
            + 'object="alert_actions.conf.spec", object_type="file"'
        )
        template = Template(
            filename=op.join(
                self._temp_obj.get_template_dir(),
                AlertActionsConfGeneration.DEFAULT_SPEC_TEMPLATE,
            )
        )
        alert_obj = Munch.fromDict(self._alert_settings)
        final_string = template.render(mod_alerts=alert_obj)
        text = linesep.join([s.strip() for s in final_string.splitlines()])
        write_file(self._alert_spec_name, self.get_spec_file_path(), text, self._logger)
        self._logger.info(
            'status="success", operation="generate", '
            + 'object="alert_actions.conf.spec", object_type="file"'
        )

    def handle(self):
        self.add_default_settings()
        # self.handler_all_icons()
        self.generate_conf()
        self.generate_spec()
        self.generate_eventtypes()
        self.generate_tags()
        # self.generate_app_conf()

    def add_default_settings(self):
        default_settings = None
        with open(self._default_settings_file) as df:
            default_settings = jloads(df.read())

        for alert in self._alert_settings:
            if ac.ALERT_PROPS not in list(alert.keys()):
                alert[ac.ALERT_PROPS] = {}
            for k, v in list(default_settings.items()):
                if k in list(alert[ac.ALERT_PROPS].keys()):
                    continue

                alert[ac.ALERT_PROPS][k] = v
                self._logger.info(
                    'status="success", operation="Add default setting", alert_name="%s", "%s"="%s"',
                    alert[ac.SHORT_NAME],
                    k,
                    v,
                )


def generate_alert_actions_conf(
    input_setting=None, package_path=None, logger=None, **kwargs
):
    obj = AlertActionsConfGeneration(
        input_setting=input_setting, package_path=package_path, logger=logger, **kwargs
    )
    obj.handle()
    return None



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import logging
import threading

import solnlib.log as log
from alert_utils.alert_utils_common.metric_collector import metric_util

from . import builder_constant

LOGS = {
    "validation": "ta_builder_validation",
    "field_extraction_validator": "ta_builder_validation",
    "app_cert_validator": "ta_builder_validation",
    "meta_manager": "ta_builder",
    "sourcetype": "ta_builder",
    "utility": "ta_builder",
    "app": "ta_builder",
    "generator": "ta_builder",
    "cim_builder": "ta_builder",
    "input_builder": "ta_builder",
    "setup_builder": "ta_builder",
    "field_extraction_builder": "ta_builder",
    "alert_builder": "ta_builder",
    "global_settings_builder": "ta_builder",
    "modular_alert": "ta_builder",
    "modular_alert_testing": "ta_builder",
    "package_add_on": "package_add_on",
}

g_log_levels = {}
g_loggers = {}
g_logger_lock = threading.Lock()

DEFAULT_LOG_LEVEL = logging.INFO

# set the context of the log
log.Logs.set_context(namespace=builder_constant.ADDON_BUILDER_APP_NAME)


@metric_util.function_run_time(tags=["tab_logger"])
def get_package_add_on_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("package_add_on"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_field_extraction_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("field_extraction_builder"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_builder_util_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("utility"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_meta_manager_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("meta_manager"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_validation_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("validation"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_field_extraction_validator_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("field_extraction_validator"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_app_cert_validator_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("app_cert_validator"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_sourcetype_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("sourcetype"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_app_instance_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("app"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_generator_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("generator"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_cim_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("cim_builder"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_input_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("input_builder"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_setup_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("setup_builder"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_modular_alert_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("modular_alert"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_modular_alert_testing_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("modular_alert_testing"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def set_log_level(level, name=None):
    if name:
        logger = _get_logger(name)
        logger.setLevel(level)
    else:
        for name in list(LOGS.values()):
            logger = _get_logger(name)
            logger.setLevel(level)


@metric_util.function_run_time(tags=["tab_logger"])
def _get_logger(name, level=logging.INFO):
    with g_logger_lock:
        l = None
        if name in g_loggers:
            l = g_loggers[name]
        else:
            l = log.Logs().get_logger(name)
            g_loggers[name] = l
        lv = g_log_levels.get(name, 10000)
        if level < lv:
            # logging.DEBUG < INFO < WARN < ERROR
            l.setLevel(level)
            g_log_levels[name] = level
        return l


@metric_util.function_run_time(tags=["tab_logger"])
def get_alert_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("alert_builder"), level)


@metric_util.function_run_time(tags=["tab_logger"])
def get_global_settings_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("global_settings_builder"), level)


HIDDEN_FIELDS = ["password", "credential", "value"]


def hide_sensitive_field(raw):
    if isinstance(raw, dict):
        cloned = dict(raw)
        for k in list(raw.keys()):
            lower_key = k.lower()
            if any([i in lower_key for i in HIDDEN_FIELDS]):
                cloned[k] = "*******"
            else:
                cloned[k] = hide_sensitive_field(cloned[k])
        return cloned
    elif isinstance(raw, list) or isinstance(raw, tuple):
        cloned = [hide_sensitive_field(i) for i in raw]
        return cloned
    return raw



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import os.path as op
import sys
from os import makedirs, remove

from splunk_add_on_ucc_framework.alert_utils.alert_utils_common.conf_parser import (
    TABConfigParser,
)

from .alert_actions_merge import merge_conf_file


def write_file(file_name, file_path, content, logger, merge="stanza_overwrite"):
    logger.debug('operation="write", object="%s" object_type="file"', file_path)

    do_merge = False
    if file_name.endswith(".conf") or file_name.endswith("conf.spec"):
        do_merge = True
    else:
        logger.info(
            'event="Will not merge file="%s", '
            + 'reason="Only support conf file merge"',
            file_path,
        )

    if file_path:
        new_file = None
        if op.exists(file_path) and do_merge:
            new_file = op.join(op.dirname(file_path), "new_" + file_name)
        if new_file:
            try:
                with open(new_file, "w+") as fhandler:
                    fhandler.write(content)
                merge_conf_file(new_file, file_path, merge)
            finally:
                if op.exists(new_file):
                    remove(new_file)
        else:
            if not op.exists(op.dirname(file_path)):
                makedirs(op.dirname(file_path))
            with open(file_path, "w+") as fhandler:
                fhandler.write(content)
            if do_merge:
                # need to process the file with conf parser
                parser = TABConfigParser()
                parser.read(file_path)
                with open(file_path, "w") as df:
                    parser.write(df)
    else:
        sys.stdout.write(
            "\n##################File {}##################\n".format(file_name)
        )
        sys.stdout.write(content)


GLOBAL_SETTING_TYPE_MAP = {"text": "text", "checkbox": "bool", "password": "password"}

GLOBAL_SETTING_VALUE_NAME_MAP = {
    "text": "content",
    "bool": "bool",
    "password": "password",
}


def convert_custom_setting(parameters):
    """
    convert
    [{
        "default_value": "message",
        "name": "notification_type",
        "required": true,
        "help_string": "Choose style of HipChat notification.",
        "possible_values": {
             "Application Card": "card",
             "Message": "message"
         },
         "label": "Notification Style",
         "format_type": "dropdownlist",
         "value": "xxxx"
      }]
    to
    [{
    "title": "customized key",
    "name": "customized name",
    "type": "text",
    "description": "description of customized key"
    }]
    """
    formated = []
    if not parameters:
        return formated

    for param in parameters:
        if param.get("format_type") not in list(GLOBAL_SETTING_TYPE_MAP.keys()):
            msg = 'format_type="{}" is not support for global setting'.format(
                param.get("format_type")
            )
            raise Exception(msg)

        one_param = {
            "title": param.get("label"),
            "name": param.get("name"),
            "type": GLOBAL_SETTING_TYPE_MAP[param.get("format_type")],
            "description": param.get("help_string"),
        }
        formated.append(one_param)

    return formated


def convert_global_setting(global_settings):
    """
      convert
      {
      "customized_settings": {
          "string_label": {
          "type": "text",
          "content": "string"
          },
          "password": {
          "type": "password",
          "password": "123"
          },
          "checkbox": {
          "type": "bool",
          "bool": true
          }
      },
      "proxy_settings": {
          "proxy_password": "sef",
          "proxy_type": "http",
          "proxy_url": "1.2.3.4",
      },
      "global_settings": {
          "log_level": "INFO"
      }
    to
      {
      "title": "Proxy",
      "name": "proxy",
      "type": "default_proxy",
      "description": "proxy settings"
      },
      {
      "title": "Account Key Title",
      "name": "username",
      "type": "default_account",
      "description": "The username of the user account"
      },
      {
      "title": "Account Secret Title",
      "name": "password",
      "type": "default_account",
      "description": "The password of the user account"
      },
      {
      "title": "customized key",
      "name": "customized name",
      "type": "text",
      "description": "description of customized key"
    """
    converted = []
    if not global_settings:
        return converted

    for type, settings in list(global_settings.items()):
        if type == "proxy_settings":
            proxy = {
                "title": "Proxy",
                "name": "proxy",
                "type": "default_proxy",
                "description": "proxy settings",
            }
            converted.append(proxy)
        elif type == "log_settings":
            logging = {
                "title": "Logging",
                "name": "logging",
                "type": "default_logging",
                "description": "logging setting",
            }
            converted.append(logging)
        elif type == "credential_settings":
            username = {
                "title": "Account Key Title",
                "name": "tab_default_account_username",
                "type": "default_account",
                "description": "The username of the user account",
            }
            password = {
                "title": "Account Secret Title",
                "name": "tab_default_account_password",
                "type": "default_account",
                "description": "The password of the user account",
            }
            converted.append(username)
            converted.append(password)
        elif type == "customized_settings":
            custom_settings = convert_custom_setting(settings)
            converted += custom_settings
    return converted


def convert_global_setting_previous(global_settings):
    """
        convert global_settings=[
            {
                "type": "proxy"
            },
            {
                "type": "logging"
            },
            {
                "type": "account"
            },
            {
                "type": "custom",
                "parameters": []
            }
        ]
            to [
        {
        "title": "Proxy",
        "name": "proxy",
        "type": "default_proxy",
        "description": "proxy settings"
        },
        {
        "title": "Account Key Title",
        "name": "username",
        "type": "default_account",
        "description": "The username of the user account"
        },
        {
        "title": "Account Secret Title",
        "name": "password",
        "type": "default_account",
        "description": "The password of the user account"
        },
        {
        "title": "customized key",
        "name": "customized name",
        "type": "text",
        "description": "description of customized key"
        }
    ]
    """
    converted = []
    if not global_settings:
        return converted

    for setting in global_settings:
        if setting.get("type") == "proxy":
            proxy = {
                "title": "Proxy",
                "name": "proxy",
                "type": "default_proxy",
                "description": "proxy settings",
            }
            converted.append(proxy)
        elif setting.get("type") == "logging":
            logging = {
                "title": "Logging",
                "name": "logging",
                "type": "default_logging",
                "description": "logging setting",
            }
            converted.append(logging)
        elif setting.get("type") == "account":
            username = {
                "title": "Account Key Title",
                "name": "username",
                "type": "default_account",
                "description": "The username of the user account",
            }
            password = {
                "title": "Account Secret Title",
                "name": "password",
                "type": "default_account",
                "description": "The password of the user account",
            }
            converted.append(username)
            converted.append(password)
        elif setting.get("type") == "custom":
            custom_settings = convert_custom_setting(setting.get("parameters"))
            converted += custom_settings
    return converted


def get_test_parameter_type(param):
    if not param:
        return None

    if param.get("format_type") in list(GLOBAL_SETTING_TYPE_MAP.keys()):
        return GLOBAL_SETTING_TYPE_MAP.get(param.get("format_type"))
    return None


def get_parameter_type(param, parameters_meta, logger):
    if not parameters_meta:
        logger.info('parameters_meta="None"')
        return None

    if not param:
        logger.info('param="None"')
        return None

    for param_meta in parameters_meta:
        if param == param_meta["name"]:
            return GLOBAL_SETTING_TYPE_MAP.get(param_meta["format_type"])

    return None


def convert_test_global_settings(test_global_settings, logger):
    """
        convert to:
    {
      "customized_settings": {
        "string_label": {
          "type": "text",
          "content": "string"
        },
        "password": {
          "type": "password",
          "password": "123"
        },
        "checkbox": {
          "type": "bool",
          "bool": true
        }
      },
      "proxy_settings": {
        "proxy_password": "sef",
        "proxy_type": "http",
        "proxy_url": "1.2.3.4",
        "proxy_rdns": "0",
        "proxy_username": "sdf",
        "proxy_port": "34",
        "proxy_enabled": "1"
      },
      "global_settings": {
        "log_level": "INFO"
      }
    }
    """
    if not test_global_settings:
        logger.info('test_global_settings="%s"', test_global_settings)
        return {}

    converted = {}
    for type, settings in list(test_global_settings.items()):
        if type == "customized_settings":
            converted["customized_settings"] = {}
            for setting in settings:
                type = get_test_parameter_type(setting)
                if not type:
                    msg = "No type for {} in customized_settings".format(setting)
                    raise NotImplementedError(msg)

                converted["customized_settings"][setting["name"]] = {
                    "type": type,
                    GLOBAL_SETTING_VALUE_NAME_MAP[type]: setting.get("value"),
                }
        elif type == "log_settings":
            converted["global_settings"] = settings
        else:
            converted[type] = settings
    return converted


def split_path(path):
    """
    split a path into a list
    """
    if not path:
        return None
    paths = []
    (head, tail) = op.split(path)
    while tail:
        paths.insert(0, tail)
        (head, tail) = op.split(head)
    (drive, rest) = op.splitdrive(head)
    if drive:
        paths.insert(0, drive)
    else:
        paths.insert(0, head)
    return paths



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# encoding = utf-8
from solnlib.splunkenv import make_splunkhome_path

ADDON_BUILDER_APP_NAME = "splunk_app_addon-builder"
ADDON_BUILDER_APP_CATEGORY = "Splunk App Add-on Builder"

# can not use common_util.make_splunk_path here. cylic import
BUILDER_WORKSPACE_ROOT = make_splunkhome_path(
    ["etc", "apps", ADDON_BUILDER_APP_NAME, "local", "builder_workspace"]
)

# cookie related keys
TA_NAME = "ta_builder_current_ta_name"
TA_DISPLAY_NAME = "ta_builder_current_ta_display_name"
TA_WIZARD_STEP = "ta_builder_wizard"
BUILT_FLAG = "built_by_tabuilder"
COOKIE_KEYS = [TA_NAME, TA_WIZARD_STEP, BUILT_FLAG, TA_DISPLAY_NAME]
COOKIE_EXPIRES_DAY = 30

# data inputs
FIELD_EXTRACTION_MI = "field_extraction_mi"
FIELD_EXTRACTION_MONITOR_MI = "_tab_internal_monitor_inputs_"
VALIDATION_MI = "validation_mi"
VALIDATION_MONITOR_MI = "_tab_internal_monitor_inputs_"

# app inspect
GLOBAL_SETTING_CONF_NAME = "settings"
APP_CERT_STANZA = "app_cert"

# global settings
USR_CREDENTIAL_SETTING = "credential_settings"
PROXY_SETTING = "proxy_settings"
LOG_SETTINGS = "log_settings"
CUSTOMIZED_SETTINGS = "customized_settings"
CREDENTIAL_SCHEMA = "default_account"
LOG_SCHEMA = "default_logging"
PROXY_SCHEMA = "default_proxy"
CUSTOMIZED_BOOL_SCHEMA = "bool"
CUSTOMIZED_TEXT_SCHEMA = "text"
CUSTOMIZED_PASSWORD_SCHEMA = "password"

CUSTOMIZED_TYPE_MAP = {"text": "text", "checkbox": "bool", "password": "password"}

# cim mapping settings
CONF_MODELS = "aob_models"
CONF_EVAL_FUNC = "aob_eval_functions"
CONF_CIM_RELATED = ("props.conf", "transforms.conf", "eventtypes.conf", "tags.conf")
CONF_FE_RELATED = ("props.conf", "transforms.conf")

TAB_KO_PREFIX = "aob_gen"

RESERVED_SOURCETYPES = ("splunk:tabuilder:validation",)



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from future import standard_library

standard_library.install_aliases()
import queue
import threading
import time

from solnlib import log

from . import metric_aggregator
from .metric_exception import MetricException

__all__ = ["NumberMetricCollector"]

logger = log.Logs().get_logger("metric_collector")


class AggregatorWorker(threading.Thread):
    def __init__(self, collector):
        super().__init__(name="AggregatorWorker")
        self.daemon = True
        self.collector = collector
        self.metric_aggregators = dict()

    def run(self):
        logger.info("Start aggregator worker.")
        stop = False
        while not stop:
            stop = self.collector.process()
        self.collector.flush_all_event()


class NumberMetricCollector:
    SUM_METRIC = "sum"
    AVG_METRIC = "avg"

    def __init__(self, event_writer):
        self.event_writer = event_writer
        self.record_queue = queue.Queue()
        self.aggregators = dict()
        self.worker = None

    def set_event_writer(self, ew):
        self.event_writer = ew

    def is_stopped(self):
        if self.worker is None:
            return True
        else:
            return not self.worker.is_alive()

    def register_metric(
        self, metric_name, metric_type=SUM_METRIC, metric_tags=[], max_time_span=10
    ):
        if metric_name in self.aggregators:
            raise MetricException("metric {} has been registered.".format(metric_name))
        if metric_type == self.AVG_METRIC:
            self.aggregators[metric_name] = metric_aggregator.NumberMetricAverage(
                metric_name, self.event_writer, metric_tags, max_time_span
            )
        elif metric_type == self.SUM_METRIC:
            self.aggregators[metric_name] = metric_aggregator.NumberMetricSum(
                metric_name, self.event_writer, metric_tags, max_time_span
            )
        else:
            raise MetricException("Metric type {} is unsupported.".format(metric_type))

    def record_metric(self, metric_name, metric_value, metric_timestamp=None):
        if metric_name not in self.aggregators:
            raise MetricException(
                "metric {} is not registered yet.".format(metric_name)
            )
        if metric_timestamp is None:
            metric_timestamp = int(time.time())
        self.record_queue.put(
            {
                "type": "metric",
                "ts": metric_timestamp,
                "v": metric_value,
                "n": metric_name,
            }
        )

    def start(self):
        if self.is_stopped() == False:
            raise RuntimeError("collector worker has been started.")
        if self.worker:
            raise RuntimeError("Worker thread is stopped, but the worker is not None.")
        self.worker = AggregatorWorker(self)
        self.worker.start()

    def stop(self, graceful=True):
        self.record_queue.put({"type": "stop"})
        if graceful:
            self.join()
        self.worker = None

    def join(self):
        if self.worker:
            self.worker.join()

    def process(self):
        """
        the entry point for the worker thread
        """
        is_stop = False
        skip_metrics = []
        record = None
        try:
            record = self.record_queue.get(timeout=1)
        except queue.Empty:
            pass
        if record:
            if record["type"] == "stop":
                is_stop = True
            else:
                metric_name = record["n"]
                self.aggregators[metric_name].aggregate_metric(record)
                skip_metrics.append(metric_name)
        self._flush_events(skip_metrics)
        return is_stop

    def _flush_events(self, skip_metrics=[]):
        ts = int(time.time())
        for k, aggr in list(self.aggregators.items()):
            if k in skip_metrics:
                continue
            aggr._flush_buckets(ts)

    def flush_all_event(self):
        for k, v in list(self.aggregators.items()):
            v.flush_all_event()



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from .base import indent, quote_string


class RestFieldBuilder:

    _kv_template = "{name} = {value}"
    _rh_template = """field.RestField(
    {name},
    required={required},
    encrypted={encrypted},
    default={default},
    validator={validator}
)"""

    def __init__(self, field, validator):
        self._field = field
        self._validator = validator

    def generate_spec(self):
        return self._kv_template.format(
            name=self._field.name,
            value="",
        )

    def _indent_validator(self):
        validator = indent(self._validator)
        return validator[4:]

    def generate_rh(self):
        return self._rh_template.format(
            name=quote_string(self._field.name),
            required=self._field.required,
            encrypted=self._field.encrypted,
            default=quote_string(self._field.default),
            validator=self._indent_validator(),
        )



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class AlertParameterInputSettingNotImplemented(Exception):
    pass


class AlertIconPathInvalid(Exception):
    pass


class AlertIconFileDoesNotExist(Exception):
    pass


class AlertActionsInValidArgs(Exception):
    pass


class AlertActionsNotSupportedFeature(Exception):
    pass


class AlertActionsDoWorkNotImplemented(Exception):
    pass


class AlertActionsFailedToLoadInputSettingJson(Exception):
    pass


class AlertTestSettingFileDoesNotExist(Exception):
    pass


class AlertTestInputUnsupportedMode(Exception):
    pass


class AlertTestJsonFileLoadFailure(Exception):
    pass


class AlertTestWritingFileFailure(Exception):
    pass


class AlertSearchCreatingFailure(Exception):
    pass


class AlertTestBuildTAFailure(Exception):
    pass


class AlertTestKillingSubprocessFailure(Exception):
    pass


class AlertTestSubprocessTimeoutFailure(Exception):
    pass


class AlertTestCodeFileNotExistFailure(Exception):
    pass


class AlertCleaningFormatFailure(Exception):
    pass



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
REST Builder.
"""


import collections

from splunktaucclib.rest_handler.schema import RestSchema

from .builder import RestBuilder, RestBuilderError

__all__ = [
    "RestBuilder",
    "RestBuilderError",
    "RestHandlerClass",
    "build",
]

__version__ = "0.0.0"


RestHandlerClass = collections.namedtuple(
    "RestHandlerClass",
    ("module", "name"),
)


def build(schema, handler, output_path, j2_env, post_process=None, *args, **kwargs):
    """
    Build REST for Add-on.

    :param schema: REST schema.
    :type schema: RestSchema
    :param handler: REST handler class import path:
            ``module.sub_module.RestHandlerClass``.
        The HandlerClass must be subclass of
        splunktaucclib.rest_handler.admin_external.AdminExternalHandler.
    :type handler: str
    :param output_path: path for output.
    :param post_process:
    :param args: args for post_process.
    :param kwargs: kwargs for post_process.
    :return:
    """

    def _parse_handler(handler_path):
        parts = handler_path.split(".")
        if len(parts) <= 1:
            raise RestBuilderError(
                "Invalid handler specified. "
                'It should be in form "module.sub_module.RestHandlerClass".'
            )
        return RestHandlerClass(
            module=".".join(parts[:-1]),
            name=parts[-1],
        )

    builder_obj = RestBuilder(schema, _parse_handler(handler), output_path)
    builder_obj.build()
    if post_process is not None:
        post_process(builder_obj, schema, *args, **kwargs)
    return builder_obj



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

SHORT_NAME = "short_name"
TITLE = "title"
PARAMETERS = "parameters"
NAME = "name"
FORMAT_TYPE = "format_type"
REQUIRED = "required"
DEFAULT_VALUE = "default_value"
LABEL = "label"
HELP_STRING = "help_string"
HELP_LINK = "help_link"
FORM_CLASS = "form_class"
FORM_STYLE = "form_style"
POSSIBLE_VALUES = "possible_values"
MODULAR_ALERTS = "modular_alerts"
PLACEHOLDER = "placeholder"
LINK_TEXT = "link_text"
LINK_URL = "link_url"
LINK_TIP = "link_tip"
LINK_TYPE = "link_type"
ICON_PATH = "icon_path"
ALERT_PROPS = "alert_props"
RESULTS = "results"
RESULT = "result"
RESULTS_FILE = "results_file"
STDIN_FIELDS = "stdin_fields"
LARGE_ICON = "largeIcon"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['splunk_add_on_ucc_framework',
 'splunk_add_on_ucc_framework.alert_utils',
 'splunk_add_on_ucc_framework.alert_utils.alert_utils_common',
 'splunk_add_on_ucc_framework.alert_utils.alert_utils_common.metric_collector',
 'splunk_add_on_ucc_framework.modular_alert_builder',
 'splunk_add_on_ucc_framework.modular_alert_builder.build_core',
 'splunk_add_on_ucc_framework.uccrestbuilder',
 'splunk_add_on_ucc_framework.uccrestbuilder.endpoint']

package_data = \
{'': ['*'],
 'splunk_add_on_ucc_framework': ['arf_dir_templates/*',
                                 'arf_dir_templates/modular_alert_package/${product_id}/appserver/static/*',
                                 'package/appserver/static/js/build/*',
                                 'package/appserver/templates/*',
                                 'package/default/*',
                                 'package/default/data/ui/nav/*',
                                 'package/default/data/ui/views/*',
                                 'schema/*',
                                 'templates/*'],
 'splunk_add_on_ucc_framework.modular_alert_builder.build_core': ['arf_template/*',
                                                                  'arf_template/default_html_theme/*']}

install_requires = \
['defusedxml>=0.7.1,<0.8.0',
 'dunamai',
 'future>=0,<1',
 'jinja2>=2,<3',
 'jsonschema>=3.2.0,<4.0.0',
 'lxml>=4.3,<5.0',
 'mako>=1,<2',
 'munch>=2,<3',
 'solnlib>=4.1.0,<5.0.0',
 'splunktaucclib>=5.0.2,<6.0.0',
 'wheel']

entry_points = \
{'console_scripts': ['install-libs = '
                     'splunk_add_on_ucc_framework:install_requirements',
                     'ucc-gen = splunk_add_on_ucc_framework:main']}

setup_kwargs = {
    'name': 'splunk-add-on-ucc-framework',
    'version': '5.5.2',
    'description': 'Splunk Add-on SDK formerly UCC is a build and code generation framework',
    'long_description': '# SPDX-FileCopyrightText: 2020 Splunk Inc.\n\n# splunk-add-on-ucc-framework\n\n![PyPI](https://img.shields.io/pypi/v/splunk-add-on-ucc-framework)\n![Python](https://img.shields.io/pypi/pyversions/splunk-add-on-ucc-framework.svg)\n\nA framework to generate UI based Splunk Add-ons. It includes UI, Rest handler, Modular input, Oauth, Alert action templates.\n\n> Note: after UCC 5.2 Python 2 specific libraries are not supported anymore.\n> This means if the add-on has `package/lib/py2/requirements.txt` they will \n> not be installed while running `ucc-gen` command. Therefore modular inputs \n> that are supposed to run on Python 2 will not be supported by UCC. \n\n## What is UCC?\n\nUCC stands for  Universal Configuration Console. It is a service for generating Splunk Add-ons which is easily customizable and flexible.\nUCC provides basic UI template for creating Addon\'s UI. It is helpful to control the activity by using hooks and other functionalities.\n\n\n## Features\n\n- Generate UCC based addons for your Splunk Technology Add-ons\n\n## UCC 5\n\nUCC 5 has potentially breaking changes to add-ons using hook extension in the UX. Previously such hooks were limited to un-optimized js files placed in the package.\nAdd-ons may now package such extensions with webpack.\n\n## Requirements\n\n- Addon package and globalConfig.json file\n\n> Note: You may refer the globalConfig.json file [here](https://github.com/splunk/addonfactory-ucc-generator/blob/master/tests/data/globalConfig.json)\n\n\n## Installation\n\n"splunk-add-on-ucc-framework" can be installed via `pip` from `PyPI`:\n\n```bash\n$ pip3 install splunk-add-on-ucc-framework\n```\n\n## How to use\n\nTo build the UCC based addon follow the below steps:\n\n1. Install the `splunk-add-on-ucc-framework` via `pip3`.\n2. Run the `ucc-gen` command.\n3. Make sure that `package` folder and `globalConfig.json` file are present in the addon folder.\n4. The final addon package will be generated, in the `output` folder.\n\n\n## Workflow\n\nBy the running the `ucc-gen` command, the following steps are executed:\n1. Cleaning out the `output` folder.\n2. Retrieve the package ID of addon.\n3. Copy UCC template directory under `output/<package_ID>` directory.\n4. Copy `globalConfig.json` file to `output/<package_ID>/appserver/static/js/build` directory.\n5. Collect and install Addon\'s requirements into `output/<package_ID>/lib` directory of addon\'s package.\n6. For the addon\'s requirements, packages are installed according to following table:\n\n    | File Name            | Description                         | Output directory in UCC build |\n    |----------------------|-------------------------------------|-------------------------------|\n    | lib/requirements.txt     | Python3 compatible packages | output/<package_ID>/lib   |\n\n7. Replace tokens in views.\n8. Copy addon\'s `package/*` to `output/<package_ID>/*` directory.\n9. If an addon requires some additional configurations in packaging than implement the steps in additional_packaging.py\n\n## Params\n\nsplunk-add-on-ucc-framework supports the following params:\n\n| Name       | Description                                                                                              |\n|------------|----------------------------------------------------------------------------------------------------------|\n| source     | Folder containing the app.manifest and app source                                                        |\n| config     | Path to the configuration file, Defaults to GlobalConfig.json in the parent directory of source provided |\n| ta-version | Optional override Current version of TA, Default version is version specified in globalConfig.json a Splunkbase compatible version of SEMVER will be used by default                         |\n',
    'author': 'rfaircloth-splunk',
    'author_email': 'rfaircloth@splunk.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/splunk/splunk-add-on-sdk-python/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__version__ = "5.0.0"

import argparse
import configparser
import glob
import json
import logging
import os
import re
import shutil
import stat
import sys
import time
from pathlib import Path

import jsonschema
from defusedxml import cElementTree as defused_et
from dunamai import Style, Version
from jinja2 import Environment, FileSystemLoader

from .app_manifest import (
    APP_MANIFEST_FILE_NAME,
    APP_MANIFEST_WEBSITE,
    AppManifest,
    AppManifestFormatException,
)
from .start_alert_build import alert_build
from .uccrestbuilder import build
from .uccrestbuilder.global_config import (
    GlobalConfigBuilderSchema,
    GlobalConfigPostProcessor,
)

sourcedir = os.path.dirname(os.path.realpath(__file__))

j2_env = Environment(loader=FileSystemLoader(os.path.join(sourcedir, "templates")))

logger = logging.getLogger("UCC")
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s [%(name)s] %(levelname)s: %(message)s")
shandler = logging.StreamHandler()
shandler.setLevel(logging.INFO)
shandler.setFormatter(formatter)
logger.addHandler(shandler)

PARENT_DIR = ".."


def get_os_path(path):
    """
    Returns a path which will be os compatible.

    Args:
        path (str): Path in string

    Return:
        string: Path which will be os compatible.
    """

    if "\\\\" in path:
        path = path.replace("\\\\", os.sep)
    else:
        path = path.replace("\\", os.sep)
    path = path.replace("/", os.sep)
    return path.strip(os.sep)


def recursive_overwrite(src, dest, ignore_list=None):
    """
    Method to copy from src to dest recursively.

    Args:
        src (str): Source of copy
        dest (str): Destination to copy
        ignore_list (list): List of files/folder to ignore while copying
    """

    if os.path.isdir(src):
        if not os.path.isdir(dest):
            os.makedirs(dest)
        files = os.listdir(src)
        for f in files:
            if not ignore_list or not os.path.join(dest, f) in ignore_list:
                recursive_overwrite(
                    os.path.join(src, f), os.path.join(dest, f), ignore_list
                )
            else:
                logger.info("Excluding : {}".format(os.path.join(dest, f)))
    else:
        if os.path.exists(dest):
            os.remove(dest)
        shutil.copy(src, dest)


def clean_before_build(outputdir):
    """
    Clean output directory before build process.

    Args:
        outputdir (str): output directory.
    """

    logger.info("Cleaning out directory " + outputdir)
    shutil.rmtree(os.path.join(outputdir), ignore_errors=True)
    os.makedirs(os.path.join(outputdir))
    logger.info("Cleaned out directory " + outputdir)


def version_tuple(version_str):
    """
    convert string into tuple to compare version

    Args:
        version_str : raw string
    Returns:
        tuple : version into tupleformat
    """
    filled = []
    for point in version_str.split("."):
        filled.append(point.zfill(8))
    return tuple(filled)


def _handle_biased_terms(conf_entities: dict) -> dict:
    for entity in conf_entities:
        entity_option = entity.get("options")
        if entity_option and "whiteList" in entity_option:
            entity_option["allowList"] = entity_option.get("whiteList")
            del entity_option["whiteList"]
        if entity_option and "blackList" in entity_option:
            entity_option["denyList"] = entity_option.get("blackList")
            del entity_option["blackList"]
    return conf_entities


def handle_biased_terms_update(schema_content: dict) -> dict:
    pages = schema_content.get("pages")
    ta_tabs = pages.get("configuration", {}).get("tabs", {})

    for tab in ta_tabs:
        conf_entities = tab.get("entity")
        tab["entity"] = _handle_biased_terms(conf_entities)

    if "inputs" in pages:
        services = pages.get("inputs", {}).get("services", {})
        for service in services:
            conf_entities = service.get("entity")
            service["entity"] = _handle_biased_terms(conf_entities)

    schema_content["meta"]["schemaVersion"] = "0.0.1"
    return schema_content


def handle_update(config_path):
    """
    handle changes in globalConfig.json

    Args:
        config_path : path to globalConfig.json

    Returns:
        dictionary : schema_content (globalConfig.json)
    """
    with open(config_path) as config_file:
        schema_content = json.load(config_file)

    version = schema_content.get("meta").get("schemaVersion", "0.0.0")

    if version_tuple(version) < version_tuple("0.0.1"):
        schema_content = handle_biased_terms_update(schema_content)
        with open(config_path, "w") as config_file:
            json.dump(schema_content, config_file, ensure_ascii=False, indent=4)

    # check for schemaVersion, if it's less than 0.0.2 then updating globalConfig.json
    if version_tuple(version) < version_tuple("0.0.2"):
        ta_tabs = schema_content.get("pages").get("configuration", {}).get("tabs", {})

        # check for schema changes in configuration page of globalConfig.json
        for tab in ta_tabs:
            if tab["name"] == "account":
                conf_entities = tab.get("entity")
                oauth_state_enabled_entity = {}
                for entity in conf_entities:
                    if entity.get("field") == "oauth_state_enabled":
                        logger.warn(
                            "oauth_state_enabled field is no longer a separate entity since UCC version 5.0.0. It is now an option in the oauth field. Please update the globalconfig.json file accordingly."
                        )
                        oauth_state_enabled_entity = entity

                    if entity.get("field") == "oauth" and not entity.get(
                        "options", {}
                    ).get("oauth_state_enabled"):
                        entity["options"]["oauth_state_enabled"] = False

                if oauth_state_enabled_entity:
                    conf_entities.remove(oauth_state_enabled_entity)

            tab_options = tab.get("options", {})
            if tab_options.get("onChange"):
                logger.error(
                    "The onChange option is no longer supported since UCC version 5.0.0. You can use custom hooks to implement these actions."
                )
                del tab_options["onChange"]
            if tab_options.get("onLoad"):
                logger.error(
                    "The onLoad option is no longer supported since UCC version 5.0.0. You can use custom hooks to implement these actions."
                )
                del tab_options["onLoad"]

        is_inputs = "inputs" in schema_content.get("pages")
        if is_inputs:
            services = schema_content.get("pages").get("inputs", {}).get("services", {})
            for service in services:
                service_options = service.get("options", {})
                if service_options.get("onChange"):
                    logger.error(
                        "The onChange option is no longer supported since UCC version 5.0.0. You can use custom hooks to implement these actions."
                    )
                    del service_options["onChange"]
                if service_options.get("onLoad"):
                    logger.error(
                        "The onLoad option is no longer supported since UCC version 5.0.0. You can use custom hooks to implement these actions."
                    )
                    del service_options["onLoad"]

        schema_content["meta"]["schemaVersion"] = "0.0.2"
        with open(config_path, "w") as config_file:
            json.dump(schema_content, config_file, ensure_ascii=False, indent=4)

    return schema_content


def replace_token(ta_name, outputdir):
    """
    Replace token with addon name in inputs.xml, configuration.xml, redirect.xml.
    Replace token with addon version in redirect.xml.

    Args:
        ta_name (str): Name of TA.
        outputdir (str): output directory.
    """

    # replace token in template
    logger.info("Replace tokens in views")
    views = ["inputs.xml", "configuration.xml", "redirect.xml"]
    for view in views:
        template_dir = os.path.join(
            outputdir, ta_name, "default", "data", "ui", "views"
        )
        with open(os.path.join(template_dir, view)) as f:
            s = f.read()

        # Safely write the changed content, if found in the file
        with open(os.path.join(template_dir, view), "w") as f:
            s = s.replace("${package.name}", ta_name)
            if view == "redirect.xml":
                s = s.replace("${ta.name}", ta_name.lower())
            f.write(s)


def install_libs(path, ucc_lib_target):
    """
    Install 3rd Party libraries in addon.

    Args:
        parent_path (str): Path of parent directory.
        ucc_lib_target (str): Target path to install libraries.
    """

    def _install_libs(requirements, ucc_target, installer="python3"):
        """
        Install 3rd Party libraries using pip2/pip3

        Args:
            requirements (str): Path to requirements file.
            ucc_target (str): Target path to install libraries.
            installer (str): Pip version(pip2/pip3).
        """
        if not os.path.exists(requirements):
            logging.warning("Unable to find requirements file. {}".format(requirements))
        else:
            if not os.path.exists(ucc_target):
                os.makedirs(ucc_target)
            install_cmd = (
                installer
                + ' -m pip install -r "'
                + requirements
                + '" --no-compile --prefer-binary --ignore-installed --use-deprecated=legacy-resolver --target "'
                + ucc_target
                + '"'
            )
            os.system(installer + " -m pip install pip --upgrade")
            os.system(install_cmd)
            remove_files(ucc_target)

    logging.info(f"  Checking for requirements in {path}")
    if os.path.exists(os.path.join(path, "lib", "requirements.txt")):
        logging.info(f"  Uses common requirements")
        _install_libs(
            requirements=os.path.join(path, "lib", "requirements.txt"),
            ucc_target=ucc_lib_target,
        )
    elif os.path.exists(
        os.path.join(os.path.abspath(os.path.join(path, os.pardir)), "requirements.txt")
    ):
        logging.info(f"  Uses common requirements")
        _install_libs(
            requirements=os.path.join(
                os.path.abspath(os.path.join(path, os.pardir)), "requirements.txt"
            ),
            ucc_target=ucc_lib_target,
        )
    else:
        logging.info(f"  Not using common requirements")

    # Prevent certain packages from being included pip could be dangerous others are just wasted space
    noshipdirs = ["setuptools", "bin", "pip", "distribute", "wheel"]
    p = Path(ucc_lib_target)
    for nsd in noshipdirs:
        try:
            # Glob can return FileNotFoundError exception if no match
            for o in p.glob(nsd + "*"):
                if o.is_dir():
                    logging.info(f"  removing directory {o} from output must not ship")
                    shutil.rmtree(o)
        except FileNotFoundError:
            pass

    # Remove execute bit from any object in lib
    NO_USER_EXEC = ~stat.S_IEXEC
    NO_GROUP_EXEC = ~stat.S_IXGRP
    NO_OTHER_EXEC = ~stat.S_IXOTH
    NO_EXEC = NO_USER_EXEC & NO_GROUP_EXEC & NO_OTHER_EXEC

    for o in p.rglob("*"):
        if not o.is_dir() and os.access(o, os.X_OK):
            logging.info(f"  fixing {o} execute bit")
            current_permissions = stat.S_IMODE(os.lstat(o).st_mode)
            os.chmod(o, current_permissions & NO_EXEC)


def remove_files(path):
    """
    Remove *.egg-info and *.dist-info files in given path.

    Args:
        path (str): Path to remove *.egg-info and *.dist-info files.
    """

    rmdirs = glob.glob(os.path.join(path, "*.egg-info")) + glob.glob(
        os.path.join(path, "*.dist-info")
    )
    for rmdir in rmdirs:
        shutil.rmtree(rmdir)


def generate_rest(ta_name, scheme, import_declare_name, outputdir):
    """
    Build REST for Add-on.

    Args:
        ta_name (str): Name of TA.
        scheme (GlobalConfigBuilderSchema): REST schema.
        import_declare_name (str): Name of import_declare_* file.
        outputdir (str): output directory.
    """

    build(
        scheme,
        "splunktaucclib.rest_handler.admin_external.AdminExternalHandler",
        os.path.join(outputdir, ta_name),
        j2_env,
        post_process=GlobalConfigPostProcessor(),
        import_declare_name=import_declare_name,
    )


def is_oauth_configured(ta_tabs):
    """
    Check if oauth is configured in globalConfig.json.

    Args:
        ta_tabs (list): List of tabs mentioned in globalConfig.json.

    Returns:
        bool: True if oauth is configured, False otherwise.
    """

    for tab in ta_tabs:
        if tab["name"] == "account":
            for elements in tab["entity"]:
                if elements["type"] == "oauth":
                    return True
            break
    return False


def replace_oauth_html_template_token(ta_name, ta_version, outputdir):
    """
    Replace tokens with addon name and version in redirect.html.

    Args:
        ta_name (str): Name of TA.
        ta_version (str): Version of TA.
        outputdir (str): output directory.
    """

    html_template_path = os.path.join(outputdir, ta_name, "appserver", "templates")
    with open(os.path.join(html_template_path, "redirect.html")) as f:
        s = f.read()

    # Safely write the changed content, if found in the file
    with open(os.path.join(html_template_path, "redirect.html"), "w") as f:
        # replace addon name in html template
        s = s.replace("${ta.name}", ta_name.lower())
        # replace addon version in html template
        s = s.replace("${ta.version}", ta_version)
        f.write(s)


def modify_and_replace_token_for_oauth_templates(
    ta_name, ta_tabs, ta_version, outputdir
):
    """
    Rename templates with respect to addon name if OAuth is configured.

    Args:
        ta_name (str): Name of TA.
        ta_version (str): Version of TA.
        ta_tabs (list): List of tabs mentioned in globalConfig.json.
        outputdir (str): output directory.
    """
    redirect_xml_src = os.path.join(
        outputdir, ta_name, "default", "data", "ui", "views", "redirect.xml"
    )
    redirect_js_src = os.path.join(
        outputdir, ta_name, "appserver", "static", "js", "build", "redirect_page.js"
    )
    redirect_html_src = os.path.join(
        outputdir, ta_name, "appserver", "templates", "redirect.html"
    )

    if is_oauth_configured(ta_tabs):
        replace_oauth_html_template_token(ta_name, ta_version, outputdir)

        redirect_js_dest = (
            os.path.join(outputdir, ta_name, "appserver", "static", "js", "build", "")
            + ta_name.lower()
            + "_redirect_page."
            + ta_version
            + ".js"
        )
        redirect_html_dest = os.path.join(
            outputdir,
            ta_name,
            "appserver",
            "templates",
            ta_name.lower() + "_redirect.html",
        )
        redirect_xml_dest = os.path.join(
            outputdir,
            ta_name,
            "default",
            "data",
            "ui",
            "views",
            ta_name.lower() + "_redirect.xml",
        )
        os.rename(redirect_js_src, redirect_js_dest)
        os.rename(redirect_html_src, redirect_html_dest)
        os.rename(redirect_xml_src, redirect_xml_dest)

    # if oauth is not configured remove the extra template
    else:
        os.remove(redirect_xml_src)
        os.remove(redirect_html_src)
        os.remove(redirect_js_src)


def add_modular_input(ta_name, schema_content, import_declare_name, outputdir):
    """
    Generate Modular input for addon.

    Args:
        ta_name (str): Name of TA.
        schema_content (dict): JSON schema of globalConfig.json.
        outputdir (str): output directory.
    """

    services = schema_content.get("pages").get("inputs").get("services")
    for service in services:
        input_name = service.get("name")
        class_name = input_name.upper()
        description = service.get("title")
        entity = service.get("entity")
        field_allow_list = ["name", "index", "sourcetype"]
        template = "input.template"
        # if the service has a template specified, use it.  Otherwise keep the default
        if "template" in service:
            template = service.get("template") + ".template"

        # filter fields in allow list
        entity = [x for x in entity if x.get("field") not in field_allow_list]
        import_declare = "import " + import_declare_name

        content = j2_env.get_template(template).render(
            import_declare=import_declare,
            input_name=input_name,
            class_name=class_name,
            description=description,
            entity=entity,
        )
        input_file_name = os.path.join(outputdir, ta_name, "bin", input_name + ".py")
        with open(input_file_name, "w") as input_file:
            input_file.write(content)

        input_default = os.path.join(outputdir, ta_name, "default", "inputs.conf")
        config = configparser.ConfigParser()
        if os.path.exists(input_default):
            config.read(input_default)

        if config.has_section(input_name):
            config[input_name]["python.version"] = "python3"
        else:
            config[input_name] = {"python.version": "python3"}

        with open(input_default, "w") as configfile:
            config.write(configfile)


def make_modular_alerts(ta_name, ta_namespace, schema_content, outputdir):
    """
    Generate the alert schema with required structure.

    Args:
        ta_name (str): Name of TA.
        ta_namespace (str): restRoot of TA.
        schema_content (dict): JSON schema of globalConfig.json.
        outputdir (str): output directory.
    """

    if schema_content.get("alerts"):
        alert_build(
            {"alerts": schema_content["alerts"]},
            ta_name,
            ta_namespace,
            outputdir,
            sourcedir,
        )


def get_ignore_list(ta_name, path):
    """
    Return path of files/folders to be removed.

    Args:
        ta_name (str): Name of TA.
        path (str): Path of '.uccignore'.

    Returns:
        list: List of paths to be removed from output directory.
    """
    if not os.path.exists(path):
        return []
    else:
        with open(path) as ignore_file:
            ignore_list = ignore_file.readlines()
        ignore_list = [
            (os.path.join("output", ta_name, get_os_path(path))).strip()
            for path in ignore_list
        ]
        return ignore_list


def remove_listed_files(ignore_list):
    """
    Return path of files/folders to removed in output folder.

    Args:
        ignore_list (list): List of files/folder to removed in output directory.

    """
    for path in ignore_list:
        if os.path.exists(path):
            if os.path.isfile(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path, ignore_errors=True)
        else:
            logger.info(
                "While ignoring the files mentioned in .uccignore {} was not found".format(
                    path
                )
            )


def update_ta_version(config, ta_version):
    """
    Update version of TA in globalConfig.json.

    Args:
        args (argparse.Namespace): Object with command-line arguments.
    """

    with open(config) as config_file:
        schema_content = json.load(config_file)
    schema_content.setdefault("meta", {})["version"] = ta_version
    with open(config, "w") as config_file:
        json.dump(schema_content, config_file, indent=4)


def handle_no_inputs(ta_name, outputdir):
    """
    Handle for configuration without input page.

    Args:
        ta_name (str): Name of TA.
        outputdir (str): output directory.
    """

    def _removeinput(path):
        """
        Remove "inputs" view from default.xml

        Args:
            path (str) : path to default.xml
        """
        tree = defused_et.parse(path)
        root = tree.getroot()

        for element in root:
            if element.tag == "view" and element.get("name") == "inputs":
                root.remove(element)

        tree.write(path)

    default_xml_file = os.path.join(
        outputdir, ta_name, "default", "data", "ui", "nav", "default.xml"
    )
    # Remove "inputs" view from default.xml
    _removeinput(default_xml_file)

    file_remove_list = []
    file_remove_list.append(
        os.path.join(outputdir, ta_name, "default", "data", "ui", "views", "inputs.xml")
    )
    file_remove_list.append(
        os.path.join(outputdir, ta_name, "appserver", "static", "css", "inputs.css")
    )
    file_remove_list.append(
        os.path.join(
            outputdir, ta_name, "appserver", "static", "css", "createInput.css"
        )
    )
    # Remove unnecessary files
    for fl in file_remove_list:
        try:
            os.remove(fl)
        except OSError:
            pass


def save_comments(outputdir, ta_name):
    """
    Save index and content of comments in conf file and return dictionary thereof
    """
    config_file = os.path.join(outputdir, ta_name, "default", "app.conf")
    comment_map = {}
    with open(config_file) as file:
        i = 0
        lines = file.readlines()
        for line in lines:
            if re.match(r"^\s*#.*?$", line):
                comment_map[i] = line
            i += 1
    return comment_map


def restore_comments(outputdir, ta_name, comment_map):
    """
    Write comments to conf file at their original indices
    """
    config_file = os.path.join(outputdir, ta_name, "default", "app.conf")
    with open(config_file) as file:
        lines = file.readlines()
    for (index, comment) in sorted(comment_map.items()):
        lines.insert(index, comment)
    with open(config_file, "w") as file:
        file.write("".join(lines))


def validate_config_against_schema(config: dict):
    """
    Validates config against JSON schema.
    Raises jsonschema.ValidationError if config is not valid.
    """
    schema_path = os.path.join(sourcedir, "schema", "schema.json")
    with open(schema_path) as f_schema:
        schema_raw = f_schema.read()
        schema = json.loads(schema_raw)
    return jsonschema.validate(instance=config, schema=schema)


def _generate(source, config, ta_version, outputdir=None):
    if outputdir is None:
        outputdir = os.path.join(os.getcwd(), "output")
    if not ta_version:
        version = Version.from_git()
        if not version.stage:
            stage = "R"
        else:
            stage = version.stage[:1]

        version_str = version.serialize(metadata=True, style=Style.SemVer)
        version_splunk = f"{version.base}{stage}{version.commit}"
        ta_version = version_splunk
    else:
        ta_version = ta_version.strip()
        version_str = ta_version

    if not os.path.exists(source):
        raise NotADirectoryError("{} not Found.".format(os.path.abspath(source)))

    # Setting default value to Config argument
    if not config:
        config = os.path.abspath(os.path.join(source, PARENT_DIR, "globalConfig.json"))

    clean_before_build(outputdir)

    app_manifest_path = os.path.abspath(
        os.path.join(source, APP_MANIFEST_FILE_NAME),
    )
    with open(app_manifest_path) as manifest_file:
        app_manifest_content = manifest_file.read()
    manifest = AppManifest()
    try:
        manifest.read(app_manifest_content)
    except app_manifest.AppManifestFormatException:
        logger.error(
            f"Manifest file @ {app_manifest_path} has invalid format.\n"
            f"Please refer to {APP_MANIFEST_WEBSITE}.\n"
            f'Lines with comments are supported if they start with "#".\n'
        )
        sys.exit(1)
    ta_name = manifest.get_addon_name()

    if os.path.exists(config):
        try:
            with open(config) as f_config:
                config_raw = f_config.read()
            validate_config_against_schema(json.loads(config_raw))
            logger.info("Config is valid")
        except jsonschema.ValidationError as e:
            logger.error("Config is not valid. Error: {}".format(e))
            sys.exit(1)

        update_ta_version(config, ta_version)

        # handle_update check schemaVersion and update globalConfig.json if required and return schema
        schema_content = handle_update(config)

        scheme = GlobalConfigBuilderSchema(schema_content, j2_env)

        ta_version = schema_content.get("meta").get("version")
        logger.info("Addon Version : " + ta_version)
        ta_tabs = schema_content.get("pages").get("configuration").get("tabs")
        ta_namespace = schema_content.get("meta").get("restRoot")
        import_declare_name = "import_declare_test"
        is_inputs = "inputs" in schema_content.get("pages")

        logger.info("Package ID is " + ta_name)

        logger.info("Copy UCC template directory")
        recursive_overwrite(
            os.path.join(sourcedir, "package"), os.path.join(outputdir, ta_name)
        )

        logger.info("Copy globalConfig to output")
        shutil.copyfile(
            config,
            os.path.join(
                outputdir,
                ta_name,
                "appserver",
                "static",
                "js",
                "build",
                "globalConfig.json",
            ),
        )
        ucc_lib_target = os.path.join(outputdir, ta_name, "lib")
        logger.info(f"Install Addon Requirements into {ucc_lib_target} from {source}")
        install_libs(source, ucc_lib_target)

        replace_token(ta_name, outputdir)

        generate_rest(ta_name, scheme, import_declare_name, outputdir)

        modify_and_replace_token_for_oauth_templates(
            ta_name, ta_tabs, schema_content.get("meta").get("version"), outputdir
        )
        if is_inputs:
            add_modular_input(ta_name, schema_content, import_declare_name, outputdir)
        else:
            handle_no_inputs(ta_name, outputdir)

        make_modular_alerts(ta_name, ta_namespace, schema_content, outputdir)

    else:
        logger.info("Addon Version : " + ta_version)
        logger.warning(
            "Skipped installing UCC required python modules as GlobalConfig.json does not exist."
        )
        logger.warning(
            "Skipped Generating UI components as GlobalConfig.json does not exist."
        )
        logger.info("Setting TA name as generic")

        ucc_lib_target = os.path.join(outputdir, ta_name, "lib")

        install_libs(source, ucc_lib_target=ucc_lib_target)

    ignore_list = get_ignore_list(
        ta_name, os.path.abspath(os.path.join(source, PARENT_DIR, ".uccignore"))
    )
    remove_listed_files(ignore_list)
    logger.info("Copy package directory ")
    recursive_overwrite(source, os.path.join(outputdir, ta_name))

    # Update app.manifest
    with open(os.path.join(outputdir, ta_name, "VERSION"), "w") as version_file:
        version_file.write(version_str)
        version_file.write("\n")
        version_file.write(ta_version)

    manifest.update_addon_version(ta_version)
    output_manifest_path = os.path.abspath(
        os.path.join(outputdir, ta_name, APP_MANIFEST_FILE_NAME)
    )
    with open(output_manifest_path, "w") as manifest_file:
        manifest_file.write(str(manifest))

    comment_map = save_comments(outputdir, ta_name)
    app_config = configparser.ConfigParser()
    app_config.read_file(open(os.path.join(outputdir, ta_name, "default", "app.conf")))
    if not "launcher" in app_config:
        app_config.add_section("launcher")
    if not "id" in app_config:
        app_config.add_section("id")
    if not "install" in app_config:
        app_config.add_section("install")
    if not "package" in app_config:
        app_config.add_section("package")
    if not "ui" in app_config:
        app_config.add_section("ui")

    app_config["launcher"]["version"] = ta_version
    app_config["launcher"]["description"] = manifest.get_description()

    app_config["id"]["version"] = ta_version

    app_config["install"]["build"] = str(int(time.time()))
    app_config["package"]["id"] = ta_name

    app_config["ui"]["label"] = manifest.get_title()

    with open(
        os.path.join(outputdir, ta_name, "default", "app.conf"), "w"
    ) as configfile:
        app_config.write(configfile)
    # restore License header
    restore_comments(outputdir, ta_name, comment_map)

    # Copy Licenses
    license_dir = os.path.abspath(os.path.join(source, PARENT_DIR, "LICENSES"))

    if os.path.exists(license_dir):
        logger.info("Copy LICENSES directory ")
        recursive_overwrite(license_dir, os.path.join(outputdir, ta_name, "LICENSES"))

    if os.path.exists(
        os.path.abspath(os.path.join(source, PARENT_DIR, "additional_packaging.py"))
    ):
        sys.path.insert(0, os.path.abspath(os.path.join(source, PARENT_DIR)))
        from additional_packaging import additional_packaging

        additional_packaging(ta_name)


def generate(source="package", config=None, ta_version=None, outputdir=None):
    _generate(source, config, ta_version, outputdir)


def main():
    parser = argparse.ArgumentParser(description="Build the add-on")
    parser.add_argument(
        "--source",
        type=str,
        nargs="?",
        help="Folder containing the app.manifest and app source",
        default="package",
    )
    parser.add_argument(
        "--config",
        type=str,
        nargs="?",
        help="Path to configuration file, defaults to GlobalConfig.json in parent directory of source provided",
        default=None,
    )
    parser.add_argument(
        "--ta-version",
        type=str,
        help="Version of TA, default version is version specified in the package such as app.manifest, app.conf, and globalConfig.json",
        default=None,
    )
    args = parser.parse_args()
    _generate(args.source, args.config, args.ta_version)



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


# encoding = utf-8

import functools
import time

from past.builtins import str
from six import string_types as str

from . import monitor

__all__ = ["initialize_metric_collector", "function_run_time"]


def initialize_metric_collector(config, update_config=False):
    """
    config is a configuration dict.
     - app: required field, define the app name for the monitor
     - event_writer: event writer type, must be one of ['file', 'hec']
     - writer_config: a dict, define the configuration for the event writer.
                     different event writer has different configurations.
    update_config: force to update the configure
    """
    m = monitor.Monitor().configure(config, force_update=update_config)
    m.start()


def write_event(ev, tags=[]):
    monitor.Monitor().write_event(ev, tags)


CREDENTIAL_KEYS = ["password", "passwords", "token"]


def mask_credentials(data):
    """
    The argument will be cloned
    """
    masked_str = "******"
    if isinstance(data, dict):
        new_data = {}
        for k in list(data.keys()):
            if isinstance(k, str):
                _key = k.lower()
                sensitive_word = False
                for w in CREDENTIAL_KEYS:
                    if w in _key:
                        sensitive_word = True
                        break
                if sensitive_word:
                    new_data[k] = masked_str
                else:
                    new_data[k] = mask_credentials(data[k])
            else:
                new_data[k] = mask_credentials(data[k])
        return new_data
    elif isinstance(data, (list, tuple)):
        new_data = []
        for d in data:
            new_data.append(mask_credentials(d))
        return new_data
    elif isinstance(data, str):
        d = data.lower()
        sensitive_word = False
        for w in CREDENTIAL_KEYS:
            if w in d:
                sensitive_word = True
                break
        if sensitive_word:
            return masked_str
    elif not isinstance(data, (int, float)):
        return "Class:" + data.__class__.__name__
    return data


max_length = 2048


def function_run_time(tags=[]):
    def apm_decorator(func):
        @functools.wraps(func)
        def func_wrappers(*args, **kwargs):
            m = monitor.Monitor()
            func_attr = {
                "function_name": func.__name__,
                "positional_args": str(mask_credentials(args))[0:max_length],
                "keyword_arguments": str(mask_credentials(kwargs))[0:max_length],
            }
            ev = {"action": "invoke"}
            ev.update(func_attr)
            m.write_event(ev, tags)
            before_invoke = int(time.time() * 1000)
            ret = func(*args, **kwargs)
            after_invoke = int(time.time() * 1000)
            ev = {"action": "done", "time_cost": (after_invoke - before_invoke)}
            ev.update(func_attr)
            m.write_event(ev, tags)
            return ret

        return func_wrappers

    return apm_decorator



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class RestmapConf:

    _admin_template = """
[admin:{namespace}]
match = /{admin_match}
members = {endpoints}
"""

    _external_template = """
[admin_external:{name}]
handlertype = python
python.version = python3
handlerfile = {rh_name}.py
handleractions = {actions}
handlerpersistentmode = true
"""

    @classmethod
    def build(cls, endpoints, namespace, admin_match):
        if not endpoints:
            return ""
        externals = [
            cls._admin_template.format(
                namespace=namespace,
                admin_match=admin_match,
                endpoints=", ".join([ep.name for ep in endpoints]),
            )
        ]
        for endpoint in endpoints:
            external = cls._external_template.format(
                name=endpoint.name,
                rh_name=endpoint.rh_name,
                actions=", ".join(endpoint.actions()),
            )
            externals.append(external)
        return "".join(externals)

    @classmethod
    def admin_externals(cls, endpoints):
        return [endpoint.name for endpoint in endpoints]


class WebConf:

    _template = """
[expose:{name}]
pattern = {name}
methods = POST, GET
"""

    _specified_template = """
[expose:{name}_specified]
pattern = {name}/*
methods = POST, GET, DELETE
"""

    _internal_template = """
[expose:{name}]
pattern = {endpoint}
methods = GET
"""

    @classmethod
    def build(cls, endpoints):
        stanzas = []
        for endpoint in endpoints:
            stanzas.append(
                cls._template.format(
                    namespace=endpoint.namespace,
                    name=endpoint.name,
                )
            )
            stanzas.append(
                cls._specified_template.format(
                    namespace=endpoint.namespace,
                    name=endpoint.name,
                )
            )
        # add splunkd data endpoint
        stanzas.append(
            cls._internal_template.format(name="_splunkd_data", endpoint="data/*")
        )
        return "".join(stanzas)



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# encode=utf-8


class NumberMetricArregator:
    """
    aggregate the metric with the the second interval.
    for minute and hour aggregate, use splunk to do it.
    """

    def __init__(self, metric_name, event_writer, metric_tags=None, max_time_span=10):
        """
        @params: metric_name - a literal string for the metric
        @params: an MetricEventWriter object which flushes the aggregated metric events
        @params: max_time_span - the aggregator creates max_time_span buckets.
                    Assume the metric event delay will not be large than max_time_span seconds.
        """
        self.m_name = metric_name
        self.m_span = max_time_span
        self.event_writer = event_writer
        self.metric_tags = metric_tags
        self.oldest_metric_time = None
        self.buckets = dict()

    def aggregate_metric(self, record):
        """
        this method is invoked in the worker thread.
        """
        if self.oldest_metric_time is None:
            self.oldest_metric_time = record["ts"]
        return self._aggregate_metric(record)

    def _aggregate_metric(self, record):
        raise NotImplemented("Children class must implement _aggregate_metric")

    def _format_metric(self, metric):
        raise NotImplemented("Children class must implement _format_metric")

    def get_formatted_metric(self, ts, metric):
        formatted_metric = self._format_metric(metric)
        formatted_metric["time_slot"] = ts
        formatted_metric["metric_name"] = self.m_name
        return formatted_metric

    def _flush_buckets(self, timestamp):
        if self.oldest_metric_time is None:
            return
        low_water_mark = timestamp - self.m_span
        if low_water_mark > self.oldest_metric_time:
            for ts in range(self.oldest_metric_time, low_water_mark):
                metric = self.buckets.get(ts, None)
                if metric:
                    self.event_writer.write_event(
                        self.get_formatted_metric(ts, metric), self.metric_tags
                    )
                    del self.buckets[ts]
            self.oldest_metric_time = low_water_mark

    def flush_all_event(self):
        for ts, metric in list(self.buckets.items()):
            self.event_writer.write_event(
                self.get_formatted_metric(ts, metric), self.metric_tags
            )


class NumberMetricSum(NumberMetricArregator):
    """
    add all the numbers in the same time slot
    """

    def __init__(self, metric_name, event_writer, metric_tags=None, max_time_span=10):
        super().__init__(metric_name, event_writer, metric_tags, max_time_span)

    def _aggregate_metric(self, record):
        ts = record["ts"]
        value = record["v"]
        bucket_value = self.buckets.get(ts, None)
        if bucket_value is None:
            self.buckets[ts] = {"sum": value}
        else:
            bucket_value["sum"] = bucket_value["sum"] + value
            self.buckets[ts] = bucket_value
        self._flush_buckets(ts)

    def _format_metric(self, metric):
        return {"sum": metric["sum"]}


class NumberMetricAverage(NumberMetricArregator):
    """
    calculate the average number for the metric
    """

    def __init__(self, metric_name, event_writer, metric_tags=None, max_time_span=10):
        super().__init__(metric_name, event_writer, metric_tags, max_time_span)

    def _format_metric(self, metric):
        return {
            "avg": float(metric["sum"]) / metric["count"],
            "max": metric["max"],
            "min": metric["min"],
        }

    def _aggregate_metric(self, record):
        ts = record["ts"]
        value = record["v"]
        bucket_value = self.buckets.get(ts, None)
        if bucket_value is None:
            self.buckets[ts] = {"sum": value, "count": 1, "max": value, "min": value}
        else:
            self.buckets[ts] = {
                "sum": bucket_value["sum"] + value,
                "count": bucket_value["count"] + 1,
                "max": value if value > bucket_value["max"] else bucket_value["max"],
                "min": value if value < bucket_value["min"] else bucket_value["min"],
            }
        self._flush_buckets(ts)



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# encoding = utf-8

import json
import logging
import time

from solnlib import log


def message(app, current_time, event, tags):
    final_event = dict(event)
    final_event["mcollector_event_ts"] = current_time
    final_event["mcollector_target_app"] = app
    if tags:
        final_event["mcollector_tags"] = tags
    return json.dumps(final_event)


class MetricEventWriter:
    def __init__(self, app, config):
        """
        config is a dict, which contains all the params for file writer
         - tag_black_list: a list of tags, if the tags are in the black list, do not write the event
         - tag_white_list: a list of tags, only write the event has this tags. white_list has higer priority than black_list
        """
        assert app is not None
        assert isinstance(config, dict)
        self._app = app
        self._deny_list_tags = config.get("tag_black_list", [])
        self._allow_list_tags = config.get("tag_white_list", [])

    def write_event(self, ev, tags=[]):
        """
        ev. event dict, it might be a hierarchy structure
        """
        assert isinstance(ev, dict)
        assert isinstance(tags, list)
        if self._allow_list_tags:
            filter_tags = [t for t in tags if t in self._allow_list_tags]
            if filter_tags:
                self._flush_event(ev, tags)
            # else: no tags found in allow_list, just skip this event
            return

        if self._deny_list_tags and tags:
            filter_tags = [t for t in tags if t in self._deny_list_tags]
            if filter_tags:
                # skip this event
                return

        self._flush_event(ev, tags)

    def _flush_event(self, ev, tags):
        ctime = int(time.time() * 1000)  # ms
        self._flush_msg(message(self._app, ctime, ev, tags))

    def _flush_msg(self, msg):
        """
        this should be implemented
        """
        raise NotImplemented("_flush_msg should be implemented.")

    def update_config(self, config):
        self._deny_list_tags = config.get("tag_black_list", [])
        self._allow_list_tags = config.get("tag_white_list", [])


class FileEventWriter(MetricEventWriter):
    def __init__(self, app, config):
        super().__init__(app, config)
        self._logger_name = config.get("logger", "default") + "_metric_events"
        self._logger = log.Logs().get_logger(self._logger_name)
        self._log_level = config.get("loglevel", "INFO")
        log.Logs().set_level(self._log_level, self._logger_name)
        # reset the formater of log handler
        for handler in self._logger.handlers:
            handler.setFormatter(logging.Formatter("%(message)s"))

    def update_config(self, config):
        super().update_config(config)
        l_name = config.get("logger", "default") + "_metric_events"
        if l_name != self._logger_name:
            self._logger_name = l_name
            self._logger = log.Logs().get_logger(self._logger_name)
        l_level = config.get("loglevel", "INFO")
        if self._log_level != l_level:
            self._log_level = l_level
            log.Logs().set_level(self._log_level, self._logger_name)

    def _flush_msg(self, msg):
        self._logger.info(msg)


class SplunkStashFileWriter(MetricEventWriter):
    """
    write a small file and use splunk rest upload this file
    """

    def __init__(self, app, config):
        super().__init__(app, config)
        pass

    def update_config(self, config):
        super(FileEventWriter, self).update_config(config)
        pass

    def _flush_msg(self, msg):
        pass



#
# Copyright 2021 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import logging
import logging.handlers
import re

import splunk.rest as rest
from splunk.util import normalizeBoolean


class InvalidResultID(Exception):
    pass


class ModularAction:
    def __init__(self, settings, logger, action_name="unknown"):
        self.settings = json.loads(settings)
        self.logger = logger
        self.session_key = self.settings.get("session_key")
        self.sid = self.settings.get("sid")
        ## if sid contains rt_scheduler with snapshot-sid; drop snapshot-sid
        ## sometimes self.sid may be an integer (1465593470.1228)
        try:
            rtsid = re.match(r"^(rt_scheduler.*)\.\d+$", self.sid)
            if rtsid:
                self.sid = rtsid.group(1)
        except:
            pass
        self.orig_sid = ""
        self.rid = ""
        self.orig_rid = ""
        self.results_file = self.settings.get("results_file")
        self.search_name = self.settings.get("search_name")
        self.app = self.settings.get("app")
        self.user = self.settings.get("user") or self.settings.get("owner")
        self.configuration = self.settings.get("configuration", {})
        ## enforce configuration is a 'dict'
        if not isinstance(self.configuration, dict):
            self.configuration = {}
        ## set loglevel to DEBUG if verbose
        if normalizeBoolean(self.configuration.get("verbose", "false")):
            self.logger.setLevel(logging.DEBUG)
            self.logger.debug("loglevel set to DEBUG")
        ## use | sendalert param.action_name=$action_name$
        self.action_name = self.configuration.get("action_name") or action_name
        ## use search_name to determine action_mode
        if self.search_name:
            self.action_mode = "saved"
        else:
            self.action_mode = "adhoc"

        self.action_status = ""
        ## Since we don't use the result object we get from settings it will be purged
        try:
            del self.settings["result"]
        except Exception:
            pass

    ## The purpose of this method is to populate the job variable with the contents from REST (/services/search/jobs/<sid>)
    ## SPL-112815 - sendalert - not all $job.<param>$ parameters come through
    def addjobinfo(self):
        self.job = {}
        if self.sid:
            try:
                response, content = rest.simpleRequest(
                    "search/jobs/%s" % self.sid,
                    sessionKey=self.session_key,
                    getargs={"output_mode": "json"},
                )
                if response.status == 200:
                    self.job = json.loads(content)["entry"][0]["content"]
                    self.logger.info(
                        self.message("Successfully retrieved search job info")
                    )
                    self.logger.debug(self.job)
                else:
                    self.logger.warn(self.message("Could not retrieve search job info"))
            except Exception as e:
                self.logger.warn(self.message("Could not retrieve search job info"))

    ## The purpose of this method is to provide a common messaging interface
    def message(self, signature, status=None):
        message = "sendmodaction - "
        message_params = {
            "signature": signature or "",
            "action_name": self.action_name or "",
            "search_name": self.search_name or "",
            "sid": self.sid or "",
            "orig_sid": self.orig_sid or "",
            "rid": self.rid or "",
            "orig_rid": self.orig_rid or "",
            "app": self.app or "",
            "user": self.user or "",
            "action_mode": self.action_mode or "",
            "action_status": status or self.action_status or "",
        }
        for k, v in message_params.items():
            # Do not include empty value params in the message.
            if v != "":
                message += f'{k}="{v}" '
        return message.rstrip()

    ## The purpose of this method is to update per-result ModAction attributes
    def update(self, result):
        ## This is for events/results that were created as the result of a previous action
        self.orig_sid = result.get("orig_sid", "")
        ## This is for events/results that were created as the result of a previous action
        self.orig_rid = result.get("orig_rid", "")
        if "rid" in result:
            self.rid = result["rid"]
        else:
            raise InvalidResultID("Result must have an ID")

    ## The purpose of this method is to generate per-result invocation messages
    def invoke(self):
        self.logger.info(self.message("Invoking modular action"))

    def dowork(self):
        return
